@@ -16,7 +16,7 @@ import sys
 from biomaj_core.utils import Utils
 from biomaj_download.downloadclient import DownloadClient
 from biomaj_download.message import downmessage_pb2
-from biomaj_download.download.http import HTTPParse
+from biomaj_download.download.curl import HTTPParse
 from biomaj_download.download.localcopy import LocalDownload
 
 from biomaj.mongo_connector import MongoConnector
@@ -24,6 +24,7 @@ from biomaj.options import Options
 from biomaj.process.processfactory import RemoveProcessFactory, PreProcessFactory, PostProcessFactory
 
 from biomaj_zipkin.zipkin import Zipkin
+from yapsy.PluginManager import PluginManager
 
 
 class Workflow(object):
@@ -334,6 +335,28 @@ class UpdateWorkflow(Workflow):
         logging.debug('New workflow')
         self.session._session['update'] = True
 
+    def _get_plugin(self, name, plugin_args):
+        options = {}
+        plugins_dir = self.bank.config.get('plugins_dir')
+        if not plugins_dir:
+            return None
+        if plugin_args:
+            for plugin_arg in plugin_args.split(','):
+                arg = plugin_arg.split('=', 1)
+                options[arg[0].strip()] = arg[1].strip()
+        requested_plugin = None
+        simplePluginManager = PluginManager()
+        simplePluginManager.setPluginPlaces([plugins_dir])
+        simplePluginManager.collectPlugins()
+        logging.error('Load plugins from %s' % (plugins_dir))
+        for pluginInfo in simplePluginManager.getAllPlugins():
+            if pluginInfo.plugin_object.name() == name:
+                requested_plugin = pluginInfo.plugin_object
+        if requested_plugin is None:
+            return None
+        requested_plugin.configure(options)
+        return requested_plugin
+
     def wf_init(self):
         err = super(UpdateWorkflow, self).wf_init()
         if not err:
@@ -431,6 +454,7 @@ class UpdateWorkflow(Workflow):
         """
         logging.info('Workflow:wf_copydepends')
         deps = self.bank.get_dependencies()
+        cf = self.session.config
         for dep in deps:
             if self.bank.config.get(dep + '.files.move'):
                 logging.info('Worflow:wf_depends:Files:Move:' + self.bank.config.get(dep + '.files.move'))
@@ -443,7 +467,10 @@ class UpdateWorkflow(Workflow):
                     logging.error('Could not find a session update for bank ' + dep)
                     return False
                 # b = self.bank.get_bank(dep, no_log=True)
-                locald = LocalDownload(bdir)
+                locald = LocalDownload(
+                    bdir,
+                    use_hardlinks=cf.get_bool("use_hardlinks", default=False)
+                )
                 (file_list, dir_list) = locald.list()
                 locald.match(self.bank.config.get(dep + '.files.move').split(), file_list, dir_list)
                 bankdepdir = self.bank.session.get_full_release_directory() + "/" + dep
@@ -484,11 +511,38 @@ class UpdateWorkflow(Workflow):
             MongoConnector.banks.update({'name': self.bank.name},
                                         info)
 
+    def __findLastRelease(self, releases):
+        '''
+        Try to find most release from releases input array
+        '''
+        release = releases[0]
+        releaseElts = re.split(r'\.|-', release)
+        logging.debug('found a release %s' % (release))
+        for rel in releases:
+            if rel == release:
+                continue
+            logging.debug('compare next release %s' % (rel))
+            relElts = re.split(r'\.|-', rel)
+            index = 0
+            for relElt in relElts:
+                logging.debug("compare release major,minor,etc. : %s >? %s" % (relElt, releaseElts[index]))
+                try:
+                    if int(relElt) > int(releaseElts[index]):
+                        release = rel
+                        logging.debug("found newer release %s" % (rel))
+                        break
+                except ValueError:
+                    pass
+                finally:
+                    index += 1
+        return release
+
     def wf_release(self):
         """
         Find current release on remote
         """
         logging.info('Workflow:wf_release')
+        release = None
         cf = self.session.config
         if cf.get('ref.release') and self.bank.depends:
             # Bank is a computed bank and we ask to set release to the same
@@ -531,6 +585,28 @@ class UpdateWorkflow(Workflow):
         self.session.previous_release = self.session.get('previous_release')
 
         logging.info('Workflow:wf_release:previous_session:' + str(self.session.previous_release))
+
+        if self.session.config.get('release.plugin'):
+            # Release already set from a previous run or an other bank
+            logging.info('Workflow:wf_release:plugin:' + str(self.session.config.get('release.plugin')))
+            plugin = self._get_plugin(self.session.config.get('release.plugin'), self.session.config.get('release.plugin_args'))
+            if plugin is None:
+                logging.error("Could not load plugin")
+                return False
+            try:
+                plugin_release = plugin.release()
+                logging.info('Workflow:wf_release:plugin:%s:%s' % (self.session.config.get('release.plugin'), plugin_release))
+                self.session.set('release', plugin_release)
+                self.session.set('remoterelease', plugin_release)
+                if self.session.previous_release == self.session.get('release') and not self.session.config.get_bool('release.control', default=False):
+                    logging.info('Workflow:wf_release:same_as_previous_session')
+                    return self.no_need_to_update()
+                else:
+                    return True
+            except Exception as e:
+                logging.exception("Plugin failed to get a release %s" % (str(e)))
+                return False
+
         if self.session.get('release'):
             # Release already set from a previous run or an other bank
             logging.info('Workflow:wf_release:session:' + str(self.session.get('release')))
@@ -603,7 +679,7 @@ class UpdateWorkflow(Workflow):
 
             save_as = None
             method = 'GET'
-            if protocol == 'directhttp' or protocol == 'directhttps' or protocol == 'directftp':
+            if protocol in ['directftp', 'directftps', 'directhttp', 'directhttps']:
                 keys = cf.get('url.params')
                 if keys is not None:
                     params = {}
@@ -631,6 +707,19 @@ class UpdateWorkflow(Workflow):
                     params['protocol'] = str(cf.get('irods.protocol')).strip()
                     params['zone'] = str(cf.get('irods.zone')).strip()
 
+            # Protocol options: as for params, a field contains the name
+            # of the options (options.names) and the values are in another
+            # field named options.<option_name>.
+            protocol_options = {}
+            option_names = cf.get('options.names')
+            if option_names is not None:
+                option_names = option_names.split(',')
+                for option_name in option_names:
+                    option_name = option_name.strip()
+                    param = cf.get('options.' + option_name)
+                    protocol_options[option_name] = param.strip()
+                logging.debug("Protocol options: " + str(protocol_options))
+
             release_downloader = dserv.get_handler(
                 protocol,
                 server,
@@ -643,10 +732,11 @@ class UpdateWorkflow(Workflow):
                 proxy_auth=proxy_auth,
                 save_as=save_as,
                 timeout_download=cf.get('timeout.download'),
-                offline_dir=self.session.get_offline_directory()
+                offline_dir=self.session.get_offline_directory(),
+                protocol_options=protocol_options
             )
 
-            if protocol == 'directhttp' or protocol == 'directhttps' or protocol == 'directftp':
+            if protocol in ['directftp', 'directftps', 'directhttp', 'directhttps']:
                 release_downloader.set_files_to_download(remotes)
             # """"""""""""""""""""""""
 
@@ -684,10 +774,13 @@ class UpdateWorkflow(Workflow):
                 tmp_dir = tempfile.mkdtemp('biomaj')
                 rel_files = release_downloader.download(tmp_dir)
                 rel_file = None
+                rel_file_name = rel_files[0]['name']
+                if 'save_as' in rel_files[0] and rel_files[0]['save_as']:
+                    rel_file_name = rel_files[0]['save_as']
                 if (sys.version_info > (3, 0)):
-                    rel_file = open(tmp_dir + '/' + rel_files[0]['name'], encoding='utf-8')
+                    rel_file = open(tmp_dir + '/' + rel_file_name, encoding='utf-8', errors='ignore')
                 else:
-                    rel_file = open(tmp_dir + '/' + rel_files[0]['name'])
+                    rel_file = open(tmp_dir + '/' + rel_file_name)
                 rel_content = rel_file.read()
                 rel_file.close()
                 shutil.rmtree(tmp_dir)
@@ -696,11 +789,18 @@ class UpdateWorkflow(Workflow):
                     logging.error('release.regexp defined but does not match any file content')
                     self._close_download_service(dserv)
                     return False
+                rels = re.findall(cf.get('release.regexp'), rel_content)
+                if len(rels) == 1:
+                    release = rels[0]
+                else:
+                    release = self.__findLastRelease(rels)
+                '''
                 # If regexp contains matching group, else take whole match
                 if len(rel.groups()) > 0:
                     release = rel.group(1)
                 else:
                     release = rel.group(0)
+                '''
 
             release_downloader.close()
             self._close_download_service(dserv)
@@ -974,6 +1074,7 @@ class UpdateWorkflow(Workflow):
         logging.info("Workflow:wf_download:DownloadSession:" + str(session))
 
         use_remote_list = False
+        use_plugin = False
 
         http_parse = HTTPParse(
             cf.get('http.parse.dir.line'),
@@ -1058,6 +1159,19 @@ class UpdateWorkflow(Workflow):
                 if cf.get('remote.file.' + str(i) + '.name'):
                     save_as = cf.get('remote.file.' + str(i) + '.name')
 
+                # Protocol options: as for params, a field contains the name
+                # of the options (options.names) and the values are in another
+                # field named options.<option_name>.
+                protocol_options = {}
+                option_names = cf.get('remote.file.' + str(i) + '.options.names')
+                if option_names is not None:
+                    option_names = option_names.split(',')
+                    for option_name in option_names:
+                        option_name = option_name.strip()
+                        param = cf.get('remote.file.' + str(i) + '.options.' + option_name)
+                        protocol_options[option_name] = param.strip()
+                    logging.debug("Protocol options: " + str(protocol_options))
+
                 subdownloader = dserv.get_handler(
                     protocol,
                     server,
@@ -1070,7 +1184,8 @@ class UpdateWorkflow(Workflow):
                     proxy_auth=proxy_auth,
                     save_as=save_as,
                     timeout_download=cf.get('timeout.download'),
-                    offline_dir=self.session.get_offline_directory()
+                    offline_dir=self.session.get_offline_directory(),
+                    protocol_options=protocol_options
                 )
                 subdownloader.set_files_to_download(remotes)
 
@@ -1094,7 +1209,7 @@ class UpdateWorkflow(Workflow):
 
             remote_dir = cf.get('remote.dir')
 
-            if protocol == 'directhttp' or protocol == 'directhttps' or protocol == 'directftp':
+            if protocol in ['directftp', 'directftps', 'directhttp', 'directhttps']:
                 keys = cf.get('url.params')
                 if keys is not None:
                     params = {}
@@ -1120,6 +1235,19 @@ class UpdateWorkflow(Workflow):
 
             save_as = cf.get('target.name')
 
+            # Protocol options: as for params, a field contains the name
+            # of the options (options.names) and the values are in another
+            # field named options.<option_name>.
+            protocol_options = {}
+            option_names = cf.get('options.names')
+            if option_names is not None:
+                option_names = option_names.split(',')
+                for option_name in option_names:
+                    option_name = option_name.strip()
+                    param = cf.get('options.' + option_name)
+                    protocol_options[option_name] = param.strip()
+                logging.debug("Protocol options: " + str(protocol_options))
+
             downloader = dserv.get_handler(
                 protocol,
                 server,
@@ -1132,10 +1260,11 @@ class UpdateWorkflow(Workflow):
                 proxy_auth=proxy_auth,
                 save_as=save_as,
                 timeout_download=cf.get('timeout.download'),
-                offline_dir=self.session.get_offline_directory()
+                offline_dir=self.session.get_offline_directory(),
+                protocol_options=protocol_options
             )
 
-            if protocol == 'directhttp' or protocol == 'directhttps' or protocol == 'directftp':
+            if protocol in ['directftp', 'directftps', 'directhttp', 'directhttps']:
                 downloader.set_files_to_download(remotes)
 
             remote_list = cf.get('remote.list', default=None)
@@ -1144,6 +1273,13 @@ class UpdateWorkflow(Workflow):
                 downloader.files_to_download = self._get_list_from_file(remote_list)
                 use_remote_list = True
 
+            remote_plugin = cf.get('remote.plugin', default=None)
+            if remote_plugin is not None:
+                logging.info("Use list from plugin %s" % (remote_plugin))
+                plugin = self._get_plugin(self.session.config.get('remote.plugin'), self.session.config.get('remote.plugin_args'))
+                downloader.files_to_download = plugin.list(self.session.get('release'))
+                use_plugin = True
+
             downloaders.append(downloader)
 
         self._close_download_service(dserv)
@@ -1160,6 +1296,10 @@ class UpdateWorkflow(Workflow):
                 if not downloader.files_to_download:
                     self.session.set('remoterelease', self.session.previous_release)
                     return self.no_need_to_update()
+            elif use_plugin:
+                if not downloader.files_to_download:
+                    self.session.set('remoterelease', self.session.previous_release)
+                    return self.no_need_to_update()
             else:
                 (file_list, dir_list) = downloader.list()
                 downloader.match(cf.get('remote.files', default='.*').split(), file_list, dir_list)
@@ -1169,6 +1309,8 @@ class UpdateWorkflow(Workflow):
                 if 'save_as' not in f or not f['save_as']:
                     f['save_as'] = f['name']
                     for p in cf.get('remote.files', default='.*').split():
+                        if p == '.*' or p == '**/*':
+                            continue
                         if p.startswith('^'):
                             p = p.replace('^', '^/')
                         else:
@@ -1183,7 +1325,6 @@ class UpdateWorkflow(Workflow):
         self.session.set('download_files', downloader.files_to_download)
         self.session._session['stats']['nb_downloaded_files'] = len(files_to_download)
         logging.info('Workflow:wf_download:nb_files_to_download:%d' % (len(files_to_download)))
-
         if self.session.get('release') and self.session.config.get_bool('release.control', default=False):
             if self.session.previous_release == self.session.get('remoterelease'):
                 if self.is_previous_release_content_identical():
@@ -1334,7 +1475,10 @@ class UpdateWorkflow(Workflow):
             logging.debug('Workflow:wf_download:Copy files from ' + last_production_dir)
             for downloader in downloaders:
                 copied_files += downloader.files_to_copy
-                Utils.copy_files(downloader.files_to_copy, offline_dir)
+                Utils.copy_files(
+                    downloader.files_to_copy, offline_dir,
+                    use_hardlinks=cf.get_bool('use_hardlinks', default=False)
+                )
 
         downloader.close()
 
@@ -1352,7 +1496,7 @@ class UpdateWorkflow(Workflow):
                 redis_prefix=self.redis_prefix
             )
             if pool_size:
-                logging.debug('Set rate limiting: %s' % (str(pool_size)))
+                logging.info('Set rate limiting: %s' % (str(pool_size)))
                 dserv.set_rate_limiting(int(pool_size))
 
         else:
@@ -1376,6 +1520,7 @@ class UpdateWorkflow(Workflow):
                 message.bank = self.name
                 message.session = session
                 message.local_dir = offline_dir
+                message.protocol_options.update(downloader.protocol_options)
                 remote_file = downmessage_pb2.DownloadFile.RemoteFile()
                 protocol = downloader.protocol
                 remote_file.protocol = downmessage_pb2.DownloadFile.Protocol.Value(protocol.upper())
@@ -1452,6 +1597,19 @@ class UpdateWorkflow(Workflow):
         logging.info("Workflow:wf_download:Download:Waiting")
         download_error = False
         try:
+            download_plugin = cf.get('download.plugin', default=None)
+            if download_plugin is not None:
+                self.downloaded_files = copied_files
+                logging.info("Use download from plugin %s" % (download_plugin))
+                plugin = self._get_plugin(self.session.config.get('download.plugin'), self.session.config.get('remote.plugin_args'))
+                if not hasattr(plugin, 'download'):
+                    logging.error('Workflow:wf_download:Plugin:%s:Error: plugin does not have a download method' % (download_plugin))
+                    return False
+                res = plugin.download(self.session.get('release'), downloader.files_to_download, offline_dir=offline_dir)
+                self._close_download_service(dserv)
+                self.downloaded_files += downloader.files_to_download
+                return res
+
             download_error = dserv.wait_for_download()
         except Exception as e:
             self._close_download_service(dserv)
@@ -1561,7 +1719,7 @@ class UpdateWorkflow(Workflow):
             self.session.get_release_directory(),
             'flat'
         )
-
+        # We use move=True so there is no need to try to use hardlinks here
         local_files = Utils.copy_files_with_regexp(from_dir, to_dir, regexp, True)
         self.session._session['files'] = local_files
         if len(self.session._session['files']) == 0:
@@ -1829,3 +1987,43 @@ class ReleaseCheckWorkflow(UpdateWorkflow):
 
     def wf_progress(self, task, status):
         return
+
+
+class RepairWorkflow(UpdateWorkflow):
+    """
+    Workflow to repaire a bank i.e. unlock bank and set all steps to OK (after a manual repair for example)
+    """
+
+    def __init__(self, bank):
+        """
+        Instantiate a new workflow
+
+        :param bank: bank on which to apply the workflow
+        :type bank: Bank
+        :param session: session to remove
+        :type session: :class:`biomaj.session.Session`
+        """
+        self.wf_unlock(bank)
+        UpdateWorkflow.__init__(self, bank)
+        self.skip_all = True
+
+    def wf_unlock(self, bank):
+        logging.info('Workflow:wf_unlock')
+        data_dir = bank.session.config.get('data.dir')
+        lock_dir = bank.session.config.get('lock.dir', default=data_dir)
+        lock_file = os.path.join(lock_dir, bank.name + '.lock')
+        maintenance_lock_file = os.path.join(lock_dir, 'biomaj.lock')
+        if os.path.exists(maintenance_lock_file):
+            logging.error('Biomaj is in maintenance')
+            return False
+        if os.path.exists(lock_file):
+            os.remove(lock_file)
+        return True
+
+    def start(self):
+        if self.session.get('release') is None:
+            logging.warn('cannot repair bank, no known release')
+            return True
+        Workflow.start(self)
+        self.wf_over()
+        return True
