@@ -21,12 +21,6 @@ from biomaj.workflow import Workflow
 from biomaj.workflow import UpdateWorkflow
 from biomaj.workflow import ReleaseCheckWorkflow
 from biomaj_core.utils import Utils
-from biomaj_download.download.ftp import FTPDownload
-from biomaj_download.download.direct import DirectFTPDownload
-from biomaj_download.download.direct import DirectHttpDownload
-from biomaj_download.download.http import HTTPDownload
-from biomaj_download.download.localcopy  import LocalDownload
-from biomaj_download.download.downloadthreads import DownloadThread
 from biomaj_core.config import BiomajConfig
 from biomaj.process.processfactory import PostProcessFactory
 from biomaj.process.processfactory import PreProcessFactory
@@ -103,7 +97,11 @@ class UtilsForTest():
       os.chmod(to_file, stat.S_IRWXU)
 
     # Manage local bank test, use bank test subdir as remote
-    properties = ['multi.properties', 'computederror.properties', 'error.properties', 'local.properties', 'localprocess.properties', 'testhttp.properties', 'computed.properties', 'computed2.properties', 'sub1.properties', 'sub2.properties']
+    properties = ['multi.properties', 'computederror.properties',
+                  'error.properties', 'local.properties',
+                  'localprocess.properties', 'testhttp.properties',
+                  'computed.properties', 'computed2.properties',
+                  'sub1.properties', 'sub2.properties']
     for prop in properties:
       from_file = os.path.join(curdir, prop)
       to_file = os.path.join(self.conf_dir, prop)
@@ -421,6 +419,58 @@ class TestBiomajFunctional(unittest.TestCase):
     b.update()
     self.assertTrue(b.session.get('update'))
 
+  def test_update_hardlinks(self):
+    """
+    Update a bank twice with hard links. Files copied from previous release
+    must be links.
+    """
+    b = Bank('local')
+    b.config.set('keep.old.version', '3')
+    b.config.set('use_hardlinks', '1')
+    # Create a file in bank dir (which is the source dir) so we can manipulate
+    # it. The pattern is taken into account by the bank configuration.
+    # Note that this file is created in the source tree so we remove it after
+    # or if this test fails in between.
+    tmp_remote_file = b.config.get('remote.dir') + 'test.safe_to_del'
+    if os.path.exists(tmp_remote_file):
+        os.remove(tmp_remote_file)
+    open(tmp_remote_file, "w")
+    # First update
+    b.update()
+    self.assertTrue(b.session.get('update'))
+    old_release = b.session.get_full_release_directory()
+    # Touch tmp_remote_file to force update. We set the date to tomorrow so we
+    # are sure that a new release will be detected.
+    tomorrow = time.time() + 3660 * 24  # 3660s for safety (leap second, etc.)
+    os.utime(tmp_remote_file, (tomorrow, tomorrow))
+    # Second update
+    try:
+        b.update()
+        self.assertTrue(b.session.get('update'))
+        new_release = b.session.get_full_release_directory()
+        # Test that files in both releases are links to the the same file.
+        # We can't use tmp_remote_file because it's the source of update and we
+        # can't use test.fasta.gz because it is uncompressed and then not the
+        # same file.
+        for f in ['test2.fasta', 'test_100.txt']:
+             file_old_release = os.path.join(old_release, 'flat', f)
+             file_new_release = os.path.join(new_release, 'flat', f)
+             try:
+                 self.assertTrue(os.path.samefile(file_old_release, file_new_release))
+             except AssertionError:
+                 msg = "In %s: copy worked but hardlinks were not used." % self.id()
+                 logging.info(msg)
+        # Test that no links are done for tmp_remote_file
+        file_old_release = os.path.join(old_release, 'flat', 'test.safe_to_del')
+        file_new_release = os.path.join(new_release, 'flat', 'test.safe_to_del')
+        self.assertFalse(os.path.samefile(file_old_release, file_new_release))
+    except Exception:
+        raise
+    finally:
+        # Remove file
+        if os.path.exists(tmp_remote_file):
+            os.remove(tmp_remote_file)
+    
   def test_fromscratch_update(self):
       """
       Try updating twice, at second time, bank should  be updated (force with fromscratc)
@@ -736,6 +786,7 @@ class TestBiomajFunctional(unittest.TestCase):
   def test_multi(self):
     b = Bank('multi')
     res = b.update()
+    self.assertTrue(res)
     with open(os.path.join(b.session.get_full_release_directory(),'flat/test1.json'), 'r') as content_file:
       content = content_file.read()
       my_json = json.loads(content)
